local nametag, addon = ...
local L = LibStub("AceLocale-3.0"):GetLocale(nametag)
-- some easy-to-manipulate constants:
local small_button_image_size,large_button_image_size = 16,24
local small_font_object,large_font_object = GameFontHighlightSmall,GameFontHighlight

addon.defaults = {
	char = {
		lastDS = "5.4.001",
	},
	profile = {
		enable = true,
		debug = 1,
		scope = 'realm',
		gvault = true,
		size = 'small',
		give_location = true,
		remind = true,
		notimeplayed = true,
		slash_rd = true,
		slash_what = true,
	},
}

addon.options = {
	--name = filled in during OnInit
	type = 'group',
	childGroups = 'tab',
	handler = addon,    -- functions listed as strings called as addon:func
	get = "GetOption",
	set = "SetOption",
	args = {
		general = {
			name = GENERAL,
			desc = L["General options"],
			type = 'group',
			order = 10,
			args = {
				version = {
					--name = filled in during OnInit
					type = 'description',
					fontSize = "large",
					cmdHidden = true,
					width = 'full',
					order = 1,
				},
				enable = {
					name = ENABLE,
					desc = L["Use this addon"],
					type = 'toggle',
					arg  = "ToggleEnable",
					order = 5,
				},
				gvault = {
					name = L["Guild vaults"],
					desc = L["Also search guild vault contents"],
					type = 'toggle',
					order = 15,
				},
				give_location = {
					name = L["Show location"],
					desc = L["Include bag and slot #; turn off if using a 'one huge bag' inventory addon"],
					type = 'toggle',
					arg  = "UpdateOptionLocals",
					order = 17,
				},
				remind = {
					name = L["Bag# Reminders"],
					desc = L["Add '<'/'>' bag direction reminders (see Help)"],
					type = 'toggle',
					arg  = "UpdateOptionLocals",
					order = 25,
				},
				break1 = {
					name = '',
					type = 'description',
					cmdHidden = true,
					width = 'full',
					order = 50,
				},
				scope = {
					name = L["Scope"],
					desc = L["How 'wide' a search should be"],
					type = 'select',
					order = 60,
					--values = filled in below
				},
				size = {
					name = L["Row Height"],
					type = 'select',
					arg  = "UpdateOptionLocals",
					order = 80,
					--values = filled in below
				},
				break2 = {
					name = '',
					type = 'description',
					cmdHidden = true,
					width = 'full',
					order = 90,
				},
				slash_rd = {
					name = L["Register /rd"],
					desc = L["Register /rd as a synonym for /rundown (changing requires /reload)"],
					type = 'toggle',
					order = 91,
				},
				slash_what = {
					name = L["Register /what"],
					desc = L["Register /what as a synonym for /rundown (changing requires /reload)"],
					type = 'toggle',
					order = 92,
				},
				notimeplayed = {
					name = L["Suppress /played login spam"],
					desc = L["Suppress DataStore's /played output during login"],
					type = 'toggle',
					order = 93,
				},
				deldata = {
					name = DELETE,
					type = 'group',
					inline = true,
					order = 100,
					args = {
						desc = {
							name = L["Use this to clear ALL DataStore-related information for a character (or guild vault)."],
							type = 'description',
							order = 1,
							width = 'full',
						},
						-- toons & delete filled in below
					},
				},
				-- sigh.  "-2" does not get placed in front of -1
				break3 = {
					name = '',
					type = 'description',
					cmdHidden = true,
					width = 'full',
					order = 300,
				},
				debug = {
					name = L["Debug"],
					desc = L["Metric crapton of output in chat"],
					type = 'select',
					arg  = "UpdateOptionLocals",
					order = 301,
				},
				main = {
					--name =  filled in locals
					type = 'execute',
					func = function()
						if addon.optionsFrame:IsVisible() then
							InterfaceOptionsFrameCancel:Click()
							HideUIPanel(GameMenuFrame)
						else
							LibStub("AceConfigDialog-3.0"):Close(nametag)
						end
						GameTooltip:Hide()
						addon:OnChatCommand()
					end,
					order = 302,
				},
			},
		},
		--profiles =   filled in during OnInit
		help = {
			name = HELP_LABEL,
			desc = L["Instructions"],
			type = 'group',
			order = -1,
			args = {},
		},
	},
}

addon.options.args.general.args.scope.values = {
	['realm'] = L["Current realm only"],
	['account'] = L["Cross-realm"],
	--['computer'] = [[Cross-account]],
}
addon.options.args.general.args.size.values = {
	['small'] = SMALL,
	['large'] = LARGE,
}
addon.options.args.general.args.debug.values = {
	[1] = OFF,
	[2] = [[Basic (toons & matches)]],
	[3] = [[Basic + UI]],
	[4] = [[Basic + slot contents]],
	[5] = ALL,
}
addon.options.args.help.args.main = addon.options.args.general.args.main

local options_button_open = L["Open Rundown"]
local options_button_back = L["Back To Rundown"]
addon.options.args.general.args.main.name = options_button_open

-----------------------------------------------------------------------------
-- other locals
local DS_expected_verstring = "6.0.001"  -- last full DS release was 4.2.002 -- prefix only
local datafail_error = [[failed getting %s data for %s, skipping - error follows:  %s]]
local type, pairs, ipairs, strsplit = _G.type, _G.pairs, _G.ipairs, _G.strsplit
local DS -- DataStore after loading
local AceGUI = LibStub("AceGUI-3.0")
--local flib = LibStub("LibFarmbuyer",true)
local new, del, copy, clear
local results, searchtargets, most_recent_blockname = {}
local get_safe_datastore_function
local can_search  -- bags (incl vault), slots (equipped), char (class info etc)
local size,remind,debug,give_location

-- We only need item names and itemlinks out of this function, not complete
-- data (mouse hovering uses SetHyperlink, feeding from client cache).  If
-- we can get such from local string extraction instead, do so.  Remember
-- if the item is not in the client cache.
local cachefail
local cache_rescan_delay = 7   -- seconds
local function GetItemInfo(x)
	local name,link
	if type(x) == 'string' then
		link = x
		name = link:match('%[([^%]]+)%]|h')
		if name then return name,link end
	end
	name,link = _G.GetItemInfo(x)
	if name then return name,link end
	cachefail = true
end

local FMT = {
	BAG_plain =    "%s %d:  %s",
	BAG_remind =   "%s <%d:  %s",
	BANK_plain =   "%s %d:  %s",
	BANK_remind =  "%s >%d:  %s",
	TOTAL_ITEM =   "%s:  " .. LIGHTYELLOW_FONT_COLOR_CODE .. "%d|r",
	TOTAL_ALL =    LIGHTYELLOW_FONT_COLOR_CODE .. L["Total Matches:  %d"] .. '|r',
}

local l_BagSlot_fmt = L["Slot %d"]
local l_Vault_coords_fmt = L["Row %d, Column %d"]
local l_BAG = _G.BAGSLOT         -- "Bag"
local l_BAGSLOT = _G.BAGSLOT
local l_BANK_STORAGE = L["Bank"] -- no globalstrings entry for 'bank'?  wtf.
local l_BANK_BAG = L["Bank Bag"] -- "Bag Slot" is not a great label, don't use globalstring here
local l_VOID_STORAGE = _G.VOID_STORAGE
local l_TAB_NUM = _G.GUILDBANK_TAB_NUMBER  -- "Tab %d"
local dsbags_to_names = {
	['0'] = _G.BACKPACK_TOOLTIP,
	['-2'] = _G.KEYRING,  -- could probably remove
	['100'] = l_BANK_STORAGE,
	['-1'] = l_BANK_STORAGE,  -- in case DS ever uses the same indexing scheme as the rest of the world
	['-3'] = _G.REAGENT_BANK,
}

-- Using "-" or "/" followed by any of these words opens the corresponding config
-- panel instead of the main window.
local slash_words = {
	["options"] = 'general', ["option"] = 'general', ["opt"] = 'general', ["opts"] = 'general', ["config"] = 'general',
	["help"] = 'help', ["?"] = 'help',
}


do
	local NONE = _G.NONE
	local selected_key, selected_blockname = NONE, NONE
	local function no_toon_selection() return (not addon.db.profile.enable) or (selected_key == NONE) end

	local charmap, guildmap, R_once

	addon.options.args.general.args.deldata.args.toons = {
		name = CALENDAR_PLAYER_NAME .. ' - ' .. FRIENDS_LIST_REALM:gsub(":.*",""),
		type = 'select',
		order = 20,
		width = 'double',
		values = function(info)
			-- This function gets called multiple times on each display!
			if R_once then return R_once end
			if not addon.db.profile.enable then return {} end -- bah

			charmap, guildmap = new(), new()
			local R = { [NONE]=NONE }
			local saved_scope = addon.db.profile.scope -- for Get*DisplayName
			addon.db.profile.scope = 'account'
			for realm in pairs(DS:GetRealms()) do
				for name, character in pairs(DS:GetCharacters(realm)) do
					local blockname = addon:GetToonDisplayName(name, character, realm)
					if blockname then
						charmap[blockname] = name .. '.' .. realm
						R[blockname] = blockname
					end
				end
				--local ret, all = get_safe_datastore_function('GetGuilds')(realm) XXX
				local all = DS:GetGuilds(realm)
				if all then
					for name, guild in pairs(all) do
						local blockname = addon:GetGuildDisplayName(name,realm)
						if blockname then
							guildmap[blockname] = name .. '.' .. realm
							R[blockname] = blockname
						end
					end
				end
			end
			addon.db.profile.scope = saved_scope
			R_once = R
			return R
		end,
		get = function(info) return selected_key end,
		set = function(info,key) selected_key = key; selected_blockname = key end,
	}

	addon.options.args.general.args.deldata.args.delete = {
		name = DELETE,
		type = 'execute',
		order = 21,
		disabled = no_toon_selection,
		func = function(info)
			local k = charmap[selected_key]
			if k then
				DS:DeleteCharacter(strsplit(".",k))
				charmap[selected_key] = nil
			else
				k = assert(guildmap[selected_key])
				guildmap[selected_key] = nil
				-- This function only exists in Containers module, so we're not removing anything extra.
				--get_safe_datastore_function('DeleteGuild')(strsplit(".",k))  XXX
				-- Surprise!  New version, now it exists everywhere.  This will
				-- delete all guild-related stuff after all, which is probably good.
				DS:DeleteGuild(strsplit(".",k))
			end
			info.handler:Printf(L["All storage data for %s|r removed."], selected_blockname)
			selected_key = NONE
			selected_blockname = NONE
			del(charmap)
			del(guildmap)
			R_once = nil
		end,
	}
end


-----------------------------------------------------------------------------
addon = LibStub("AceAddon-3.0"):NewAddon(addon, nametag,
		        "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")

-- Its on-disk name and the global table name are the same, which simplifies things.
function addon:DS_sanity_check (modulename)
	if not _G[modulename] then return end
	-- There is also a DataStore.Version string, which is a different string
	-- format and stores a different number.  Possibly it represents an internal
	-- storage version?  Avoid for now and stick with the number that players
	-- can see during downloads.
	local ver = GetAddOnMetadata(modulename, "Version") or "?"
	-- force this to either true or nil
	return ver:match('^'..DS_expected_verstring) and true
end

-- Thanks to jerry for the nifty arg idea.
function addon:SetOption (info, value)
	local name = info[#info]
	self.db.profile[name] = value
	local arg = info.arg
	if arg then self[arg](self) end
end

function addon:GetOption (info)
	local name = info[#info]
	return self.db.profile[name]
end

-- Be aware that everything in ... still gets evaluated even when not debugging.
function addon:DPrint (d,s,...)
	if d > debug then return end
	if d==3 and debug==4 then return end -- FIXME use some proper bitflags you tard
	if s:find('%',nil,true) then
		self:Print(s:format(...))
	else
		self:Print(s,...)
	end
end

function addon:OnInitialize()
	self.pretty_verstring = nametag .. " |cff30adff" .. (GetAddOnMetadata(nametag, "Version") or "?") .. "|r"
	self.db = LibStub("AceDB-3.0"):New("RundownIVDB", self.defaults, --[[Default=]]true)

	local AceDBOptions = LibStub("AceDBOptions-3.0", true)
	if AceDBOptions then
		self.options.args.profiles = AceDBOptions:GetOptionsTable(self.db)
		self.options.args.profiles.order = 200
	end

	self.options.name = GetAddOnMetadata(nametag, "Title")
	self.options.args.general.args.version.name = self.pretty_verstring

	-- This only works because of the initialize->enable behavior of aceaddon.
	-- DS_Char will always be in front of us in either queue.
	-- <DS_Char New> -> <R4 New> -> <DS_Char OnInit> -> <R4 OnInit> -> <DS_Char OnEnable> -> <R4 OnEnable>
	if self.db.profile.notimeplayed then
		self.old_RTP = _G.RequestTimePlayed
		_G.RequestTimePlayed = function() end
	end

	self:SetEnabledState(self.db.profile.enable)
	self:UpdateOptionLocals()
	--self:RegisterEvent("ADDON_LOADED")

	-- This is kind of a kludge.
	if DataStoreHelp_Text then
		local s = DataStoreHelp_Text:GetText()
		local s2,n = s:gsub(
			"At this point, characters and guilds can be erased from Altoholic's UI.",
			"%0  |cffffffffRundown IV|r users can erase specific data from its General options."
		)
		if n and n>0 then
			DataStoreHelp_Text:SetText(s2)
		end
	end

	self.OnInitialize = nil
end


function addon:OnEnable()
	if self.old_RTP then
		_G.RequestTimePlayed = self.old_RTP
		self.old_RTP = nil
	end
	self:RegisterEvent("PLAYER_LOGOUT")
	if not self.db.profile.enable then
		return self:Disable()
	end
	--self:test_lods()
	if not self.DS_sanity_check then
		-- have passed this way before
		return
	end

	can_search = {
		-- char modules is not strictly speaking part of what gets searched
		char = self:DS_sanity_check('DataStore_Characters'),
		bags = self:DS_sanity_check('DataStore_Containers'),
		slots = self:DS_sanity_check('DataStore_Inventory'),
	}
	local sane
	for k,v in next, can_search do
		sane = sane or v
	end
	if not sane then
		self:Print(self.error_exit)
		return self:Disable()
	end

	-- Some things we'd like to do only once, but doing them in OnInit would
	-- be too early.
	LibStub("AceConfig-3.0"):RegisterOptionsTable(nametag, self.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(nametag, "Rundown")
	self.optionsFrame.okay = function() 
		addon.options.args.general.args.main.name = options_button_open
	end
	self.optionsFrame.refresh = self.optionsFrame.okay
	self.optionsFrame.cancel = self.optionsFrame.okay

	self.error_exit = nil
	self.DS_sanity_check = nil
	DS = _G.DataStore
	self:RegisterChatCommand("rundown", "OnChatCommand")
	-- Instead of creating multiple copies of the same tiny wrapper, abuse knowledge
	-- of implementation to get a reference to the created wrapper, and just reuse
	-- that as the handler.
	local slashfunc = _G.SlashCmdList[ LibStub("AceConsole-3.0").commands["rundown"] ]
	assert(type(slashfunc)=='function')
	if self.db.profile.slash_rd then
		self:RegisterChatCommand("rd", slashfunc)
	end
	if self.db.profile.slash_what then
		self:RegisterChatCommand("what", slashfunc)
	end
	table.insert(UISpecialFrames,nametag)

	-- transition hacks
	local lastDS, changed = self.db.char.lastDS, false
	if lastDS ~= DS_expected_verstring then
		if lastDS == "5.4.001" then
			-- remove old void storage stuff
			-- worked for 3 years but now DS actually supports it
			if can_search.bags and
			   _G.DataStore_Containers.ThisCharacter.Containers.Bag1216
			then
				_G.DataStore_Containers.ThisCharacter.Containers.Bag1216 = nil
				self:Printf(L["You will need to revisit %s."], VOID_STORAGE)
				changed = true
			end
		end
		self.db.char.lastDS = DS_expected_verstring
	end
	if changed then
		self:Printf(L["Updated from stored version %s to version %s"], lastDS, DS_expected_verstring)
	end
end

function addon:OnDisable()
end

function addon:ToggleEnable()
	if self.db.profile.enable then
		self:Enable()
	else
		self:Disable()
	end
end

-- locals are good
function addon:UpdateOptionLocals()
	size = self.db.profile.size
	remind = self.db.profile.remind
	debug = self.db.profile.debug
	give_location = self.db.profile.give_location
end

function addon:OnChatCommand (input)
	if input then
		input = input:trim()
		local p,word = input:sub(1,1), input:sub(2)
		if p == '-' or p == '/' then
			word = slash_words[word]
			if word then
				LibStub("AceConfigDialog-3.0"):Open(nametag)
				LibStub("AceConfigDialog-3.0"):SelectGroup(nametag,word)
				return
			end
		end
	end

	if (not self.display) or (not self.display:IsShown()) then
		self:BuildWindow()
	end

	if (not input) or (#input == 0) then
		self.display:SetText(L["Type or drag an item here"])
		self.display.editbox:HighlightText()
		return
	end
	self:DoSearch(input)
end

function addon:PLAYER_LOGOUT()
	self.db.profile.debug = nil
end

-- Tables for holding results are reused until the player gets into a fight.
function addon:PLAYER_REGEN_DISABLED()
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	if self.display and self.display:IsShown() then
		self.display:Hide()
	end
	most_recent_blockname = nil
	wipe(results)
	clear()
end

--[====[do
	local fixups = {
		-- Always register this hook even if the toon cannot use void storage
		-- right now; they may purchase the unlocking fee and we want to be
		-- able to update right away.
		['Blizzard_VoidStorageUI'] =
			function (self)
				if can_search.bags then self:VoidStorageKludges() end
			end,
	}
	local count = 0; for _ in next, fixups do count = count + 1 end

	function addon:ADDON_LOADED (_, modname)
		local f = fixups[modname]
		if f then
			fixups[modname] = nil
			count = count - 1
			if count < 1 then
				self:UnregisterEvent("ADDON_LOADED")
				fixups = nil
			end
			f(self)
		end
	end
	function addon:test_lods()
		for mod in next, fixups do
			if IsAddOnLoaded(mod) then
				self:ADDON_LOADED(nil,mod)
			end
		end
		self.test_lods = nil
	end
end]====]

function addon:find (needle, haystack)
	-- go go gadget super-defensive coding
	if needle == nil or haystack == nil then return end
	haystack = tostring(haystack):lower()

	-- hardcode to plain
	return haystack:find(needle,nil,true)
end


-----------------------------------------------------------------------------
--[===[
A vault tab is 7 rows high and 7 "columns" wide, with each "column" being two
actual columns of buttons:
	--Col1-- --Col2--   Slot  1  ==>  C1B1
	|B1||B8| |B1||B8|   Slot  2  ==>  C1B2
	|B2||B9| |B2||B9|   Slot  8  ==>  C1B8
	|B3||B10   ...      Slot 15  ==>  C2B1
	|B4||B11
	|B5||B12
	|B6||B13
	|B7||B14
The relevent constant is NUM_SLOTS_PER_GUILDBANK_GROUP = 14, but it's not
worth {loading,waiting for} Blizzard_GuildBankUI just to get that number.
If it changes, it'll be big news anyhow.

As we use this routine elsewhere, we can't assume the 14 either...
]===]
local function slot2coords (slot, percol)
	local index, col, row   -- index into a "column"
	index = slot % percol
	if index == 0 then
		index = percol
	end
	col = 2 * math.ceil((slot-0.5)/percol)

	local halfheight = percol/2
	-- row = index % vault_height (and 0->vault_height)
	if index <= halfheight then
		row = index
		col = col - 1
	else
		row = index - halfheight
	end

	return row, col
end


function addon:add_to_results (blockname, whence, itemlink, ...)
	self:DPrint(2, "***", blockname, whence, itemlink, ...)

	local entry
	if blockname ~= most_recent_blockname then
		entry = new()
		entry.name = blockname
		--entry.whence = whence -- needed?
		table.insert(results,entry)
		most_recent_blockname = blockname
	end

	-- these last two may be nil
	local cID, slot, count = ...
	cID = tostring(cID)
	slot = slot and tonumber(slot)

	entry = new()
	entry.itemlink = itemlink
	entry.whence = whence
	entry.count = count and tonumber(count)

	-- I was really expecting more here.  Possbly merge with :DoSearch.

	if whence == 'slots' then
		-- nothing else to do...?

	elseif whence == 'vault' then
		entry.tabname, entry.tabicon = select(4,...)
		if slot and give_location then
			local row, col = slot2coords(slot,14)
			entry.slot = l_Vault_coords_fmt:format(row,col)
		end

	elseif whence == 'bags' then
		entry.baglink, entry.bagicon = select(4,...)
		-- 'Bag' here is a hardcoded prefix in DS_Containers, unlikely (?) to ever
		-- change.  There is a wrapper that takes 'X' and returns the BagX object, but
		-- the corresponding get-all-objects function has handed back a table over
		-- which we iterate with pairs(), so we only ever have the BagX index string
		-- rather than 'X'.
		if cID:sub(1,3) == 'Bag' then
			local dsbag,s_dsbag
			s_dsbag = cID:sub(4); dsbag = tonumber(s_dsbag) or 0
			if dsbag > 0 and dsbag <= NUM_BAG_SLOTS then
				-- 'Bag 2: [blah]'
				entry.baglink = give_location
				                and FMT[remind and 'BAG_remind' or 'BAG_plain']:format(l_BAGSLOT,dsbag,entry.baglink)
				                or l_BAG

			elseif dsbag > NUM_BAG_SLOTS and dsbag <= NUM_BAG_SLOTS+NUM_BANKBAGSLOTS then
				-- 'Bank Bag 6: [blah]'
				entry.baglink = give_location
				                and FMT[remind and 'BANK_remind' or 'BANK_plain']:format(l_BANK_BAG,dsbag-NUM_BAG_SLOTS,entry.baglink)
				                or l_BANK_BAG

			elseif dsbag == REAGENTBANK_CONTAINER then
				-- 'Reagent Bank: [blah]'
				-- the icon is the same as what DS returns as a special case for its
				-- getcontainerinfo call, but it isn't actually stored in the container
				-- icon field, so we fake it also.  Coordinates here work like
				-- a vault tab.
				entry.baglink = dsbags_to_names[s_dsbag]
				entry.bagicon = [[Interface\ICONS\INV_Misc_Bag_SatchelofCenarius]]
				local row, col = slot2coords(slot,14)
				entry.slot = l_Vault_coords_fmt:format(row,col)
				slot = nil    -- fool the test about thirty lines from now

			else
				entry.baglink = dsbags_to_names[s_dsbag]

			end

		elseif cID:sub(1,15) == 'VoidStorage.Tab' then
			local vstab = tonumber(cID:sub(16)) or 0
			-- We need to change the slot entry to be coordinates instead of a
			-- simple slot integer.  There is no bag link/icon so those need
			-- useful values as well (DS' ScanVoidstorage only stores IDs, so
			-- we use the same link that we used to fake ourselves).
			entry.baglink = l_VOID_STORAGE .. ' ' .. l_TAB_NUM:format(vstab)
			-- The icons have changed since then.  We could, potentially, look
			-- these up programmatically, if Blizzard_VoidStorageUI is loaded...
			if vstab == 1 then
				entry.bagicon = [[Interface\ICONS\INV_Enchant_EssenceCosmicGreater]]
			elseif vstab == 2 then
				entry.bagicon = [[Interface\ICONS\INV_Enchant_EssenceArcaneLarge]]
			else
				-- ??? use our old one
				entry.bagicon = [[Interface\ICONS\Spell_Nature_AstralRecalGroup]]  -- that's right, "Recal"
			end
			if slot and give_location then
				-- Void storage grid works the same as the 'vault' case above, except
				-- it's 8 rows tall by 5 double-wide columns.  These numbers are
				-- hardcoded in Blizzard_VoidStorageUI with no named constants.
				local row, col = slot2coords(slot,16)
				entry.slot = l_Vault_coords_fmt:format(row,col)
				slot = nil    -- fool the test about ten lines from now
			end

		else
			-- I don't think we can usefully figure out the container ID, if datastore hasn't already done so.
		end
		if slot and give_location then
			entry.slot = l_BagSlot_fmt:format(slot)
		end

	end

	table.insert(results,entry)
end


-- In all these following functions, the CHAR parameter is a DS character object.

-- DataStore information can be remarkably patchy and full of gaps.  If toon
-- X has been seen by (e.g.) DS_Containers, and then DS_Characters is installed,
-- but X has not yet logged on, then toon Y will get X as part of the loops
-- below -- except that DS_Characters methods will fail with nil errors due to
-- the lack of checking.  Possibly this never happens to normal users.  Maybe
-- all the time.  Who knows.  To be fair, I'm not sure what I would do either.
--
-- As an attempt to avoid this, we prod modules with one of their supplied
-- methods, as it becomes convenient.  If any of them fail, we exclude that
-- character from the loops.
-- XXX should notify user to login to that toon & examine bank
do
	local xpcall = _G.xpcall
	local function err(...) return ... end
	local safe_functions = {}
	function get_safe_datastore_function(ds_func_name)
		if not safe_functions[ds_func_name] then
			safe_functions[ds_func_name] = function (a1,a2)
				local function f() return DS[ds_func_name](DS,a1,a2) end
				return xpcall (f, err)
			end
		end
		return safe_functions[ds_func_name]
	end
end

-- Returns a name appropriate for displaying to the user, and also as a
-- unique label for subsequent searches.  Returns nil if this character
-- should be avoided.
function addon:GetToonDisplayName (fallback, char, realm)
	local ret, nstring
	local scope = self.db.profile.scope

	if can_search.char then
		ret, nstring = get_safe_datastore_function('GetColoredCharacterName')(char)
		if not ret then
			self:DPrint(2, datafail_error, '<Character module>', tostring(char), nstring)
			return
		end
		if scope == 'realm' then           -- name is sufficient
			nstring = nstring .. '|r'
		elseif scope == 'account' then     -- use name-realm
			nstring = nstring .. ' - ' .. realm .. '|r'
		elseif scope == 'computer' then    -- for now, only name-realm is useful
			nstring = nstring .. ' - ' .. realm .. '|r'
		end
	else
		nstring = fallback -- really? that's all we can manage? crumbs.
	end

	return nstring
end
function addon:GetGuildDisplayName (guildname, realmname)
	-- Less fragile than for characters.  Multiple string concats in a single
	-- statement are faster than string.format when the variables are already
	-- of string type.
	if realmname then
		return ORANGE_FONT_COLOR_CODE .. '<' .. guildname .. '> - ' .. realmname .. '|r'
	else
		return ORANGE_FONT_COLOR_CODE .. '<' .. guildname .. '>|r'
	end
end

function addon:SearchCharacterSlots (blockname, char, texts)
	self:DPrint(2, ">>> starting :SearchCharacterSlots", blockname)
	local ret, inv = get_safe_datastore_function('GetInventory')(char)
	if not ret then
		self:DPrint(2, datafail_error, '<Inventory module>', tostring(char), inv)
		return
	end

	-- These entries might be itemlinks (if DS thinks they're enchanted) or plain
	-- itemIDs.  Either can safely be fed to GetItemInfo, which will fill out any
	-- missing bits and give a complete link.
	for i,text in ipairs(texts) do
		-- ipairs(inv) stops on the first empty slot, so we go a little more clunky
		for slot = 1, INVSLOT_LAST_EQUIPPED do while true do
			local item = inv[slot]
			if not item then break end   -- fake a continue
			local itemname, itemlink = GetItemInfo(item)
			self:DPrint(4, "*inv", text, slot, item, itemlink)
			if self:find (text, itemname) then
				self:add_to_results (blockname, 'slots', itemlink)
			end
			break
		end end
	end

	self:DPrint(2, "<<< done with :SearchCharacterSlots", blockname)
end


function addon:SearchContainer (blockname, text, containerID, container, is_vault)
	self:DPrint(4, "*cnt", containerID)
	if type(container.ids) ~= 'table' then return end  -- empty bag
	if type(container.size) ~= 'number' then return end  -- empty guild vault tab
	for slot = 1, container.size do while true do
		local itemid, itemname, itemlink
		itemid = container.ids[slot]
		if not itemid then break end  -- empty slot
		itemlink = container.links and container.links[slot] 
		-- that itemlink might be nil if there's nothing special about that item
		itemname, itemlink = GetItemInfo(itemlink or itemid)
		self:DPrint(4, "*cntslot", slot, itemid, itemname, itemlink)
		if self:find (text, itemname) then
			if is_vault then
				self:add_to_results (blockname, 'vault', itemlink,
					containerID, slot, container.counts and container.counts[slot],
					container.name, container.icon)
			else
				self:add_to_results (blockname, 'bags', itemlink,
					containerID, slot, container.counts and container.counts[slot],
					container.link, container.icon)
			end
		end
		break
	end end
end


-- The guild argument from DS routines has changed from the actual guild to
-- only the guild key instead.
function addon:SearchGuildVault (blockname, guild_key, texts)
	self:DPrint(2, ">>> starting :SearchGuildVault", blockname
	--[===[@debug@
	, #DataStore_Containers.db.global.Guilds[guild_key].Tabs
	--@end-debug@]===]
	)

	-- No way to query DS for the total number of tabs in a guild vault (see
	-- ticket #25).  No way to detect it without doing nasty things.  Instead
	-- we loop over all possible vault tabs.  The :Get routine always returns
	-- a table so the only way to tell whether it's an actual valid tab is to
	-- hackishly test for a non-nil 'name' field.
	for i,text in ipairs(texts) do
		for tabid = 1, MAX_GUILDBANK_TABS do
			local tab = DS:GetGuildBankTab(guild_key,tabid)
			if tab.name then -- avoid needlessly calling a function only to return instantly
				self:SearchContainer (blockname, text, tabid, tab, true)
			end
		end
	end

	self:DPrint(2, "<<< done with :SearchGuildVault", blockname)
end


function addon:SearchCharacterBags (blockname, char, texts)
	self:DPrint(2, ">>> starting :SearchCharacterBags", blockname)
	local ret, all = get_safe_datastore_function('GetContainers')(char)
	if not ret then
		self:DPrint(2, datafail_error, '<Containers module>', tostring(char), all)
		return
	end

	for i,text in ipairs(texts) do
		for bagid, bagdata in pairs(all) do
			self:SearchContainer (blockname, text, bagid, bagdata, false)
		end
	end

	self:DPrint(2, "<<< done with :SearchCharacterBags", blockname)
end


function addon:SearchCharacter (blockname, char, texts)
	self:DPrint(2, ">> starting :SearchCharacter", blockname)

	if can_search.slots then
		self:SearchCharacterSlots (blockname, char, texts)
	end

	if can_search.bags then
		self:SearchCharacterBags (blockname, char, texts)
	end

	self:DPrint(2, "<< done with :SearchCharacter", blockname)
end


function addon:Search (texts)
	self:DPrint(2, "> starting :Search")
	cachefail = nil
	most_recent_blockname = nil
	while #results > 0 do
		del(table.remove(results))
	end

	local scope = self.db.profile.scope
	assert(scope=='realm' or scope=='account' or scope=='computer', L["ERROR:  Scope settings are broken.  (How did you manage that?)  You should open the options panel and reset the current profile to default settings."])

	local do_vaults = self.db.profile.gvault

	if scope == 'realm' then
		self:DPrint(2, "Scope:  cross-toons")
		for name, character in pairs(DS:GetCharacters()) do
			local blockname = self:GetToonDisplayName(name, character)
			if blockname then
				self:SearchCharacter (blockname, character, texts)
			end
		end
		if can_search.bags and do_vaults then
			--local ret, all = get_safe_datastore_function('GetGuilds')() XXX
			local all = DS:GetGuilds()
			if all then
				for name, guild in pairs(all) do
					self:SearchGuildVault (self:GetGuildDisplayName(name), guild, texts)
				end
			end
		end

	elseif scope == 'account' then
		self:DPrint(2, "Scope:  cross-realm")
		for realm in pairs(DS:GetRealms()) do
			for name, character in pairs(DS:GetCharacters(realm)) do
				local blockname = self:GetToonDisplayName(name, character, realm)
				if blockname then
					self:SearchCharacter (blockname, character, texts)
				end
			end
			if can_search.bags and do_vaults then
				--local ret, all = get_safe_datastore_function('GetGuilds')(realm) XXX
				local all = DS:GetGuilds(realm)
				if all then
					for name, guild in pairs(all) do
						self:SearchGuildVault (self:GetGuildDisplayName(name,realm), guild, texts)
					end
				end
			end
		end

	--[====[
	elseif scope == 'computer' then
		-- This is largely useless until we (and DS (and Blizzard)) figure out
		-- a way of labelling an account beyond "Default".
		self:DPrint(2, "Scope:  cross-account")
		for account in pairs(DS:GetAccounts()) do
			for realm in pairs(DS:GetRealms(account)) do
				for name, character in pairs(DS:GetCharacters(realm,account)) do
					local blockname = self:GetToonDisplayName(name, character, realm)
					if blockname then
						self:SearchCharacter (blockname, character, texts)
					end
				end
				if do_vaults then
					local ret, all = get_safe_datastore_function('GetGuilds')(realm,account)
					if all then
						for name, guild in pairs(all) do
							self:SearchGuildVault (self:GetGuildDisplayName(name,realm), guild, texts)
						end
					end
				end
			end
		end]====]
	end

	self:DPrint(2, "< done with :Search")
end


-----------------------------------------------------------------------------
local function BlocklineHeading_OnRelease (widget)
	widget.label:ClearAllPoints()
	widget.label:SetPoint("TOP")
	widget.label:SetPoint("BOTTOM")
	widget.OnRelease = nil
end

local function mkblockline (text)
	local line = AceGUI:Create("Heading")
	line:SetText(text)
	line:SetFullWidth(true)
	line:SetHeight(30)
	-- No API exists for this, and it takes forever getting them to add one.
	-- Just do it ourselves and put things back when we're done.
	-- Also, :SetJustifyH("LEFT") doesn't do a #$%#$%#%#% thing no matter
	-- what actions it's combined with.
	line.OnRelease = BlocklineHeading_OnRelease
	line.label:ClearAllPoints()
	line.label:SetPoint("LEFT", 25, 0)
	return line
end

local function LineLabel_OnClick (widget)
	if IsModifiedClick("CHATLINK") and ChatEdit_GetActiveWindow() then
		-- totally cheating here:  get the first widget in the row and pull out its text
		ChatEdit_InsertLink(widget.parent.children[1].label:GetText())
	end
end
local function LineLabel_OnEnter (widget)
	if widget:GetUserData('isitem') then
		GameTooltip:SetOwner (widget.frame, "ANCHOR_LEFT", -20, 0)
		GameTooltip:SetHyperlink (widget.label:GetText())
	end
	--widget.parent:SetHighlight(1.0,0.9,0.0,0.5)  -- lib-st defaults, yellowish
	widget.parent:SetHighlight (0.3, 0.3, 0.3, 0.5)
end
local function LineLabel_OnLeave (widget)
	widget.parent:SetHighlight(0,0,0,0)
	GameTooltip:Hide()
end

local function mkitemlabel (itemlink, count)
	local itemlabel = AceGUI:Create("InteractiveCountingLabel")
	itemlabel:SetUserData('isitem',true)
	if size == 'small' then
		--itemlabel.count:SetFontObject(NumberFontNormalSmall)
		itemlabel:SetFontObject(small_font_object)
		itemlabel:SetImageSize(small_button_image_size,small_button_image_size)
		if count and count > 1 then
			itemlabel:SetText(itemlink .. ' x' .. count)
		else
			itemlabel:SetText(itemlink)
		end
	elseif size == 'large' then
		itemlabel.count:SetFontObject(NumberFontNormal)
		itemlabel:SetFontObject(large_font_object)
		itemlabel:SetImageSize(large_button_image_size,large_button_image_size)
		itemlabel:SetText(itemlink)
		itemlabel:SetCountText(count)
	end
	itemlabel:SetRelativeWidth(0.4)
	itemlabel:SetCallback("OnClick", LineLabel_OnClick)
	itemlabel:SetCallback("OnEnter", LineLabel_OnEnter)
	itemlabel:SetCallback("OnLeave", LineLabel_OnLeave)
	local id = itemlink:match("item:(%d+)")
	itemlabel:SetImage(GetItemIcon(id))
	return itemlabel
end

local function mklocationlabel (text, iconpath)
	local loclabel = AceGUI:Create("InteractiveCountingLabel")
	if size == 'small' then
		loclabel:SetFontObject(small_font_object)
		loclabel:SetImageSize(small_button_image_size,small_button_image_size)
	elseif size == 'large' then
		loclabel:SetFontObject(large_font_object)
		loclabel:SetImageSize(large_button_image_size,large_button_image_size)
	end
	loclabel:SetRelativeWidth(0.4)
	loclabel:SetCallback("OnClick", LineLabel_OnClick)
	loclabel:SetCallback("OnEnter", LineLabel_OnEnter)
	loclabel:SetCallback("OnLeave", LineLabel_OnLeave)
	loclabel:SetText(text)
	loclabel:SetImage(iconpath)
	return loclabel
end

local function mkslotline (itemlink)
	local itemlabel = mkitemlabel(itemlink)
	local eq = mklocationlabel(CURRENTLY_EQUIPPED)
	eq:SetRelativeWidth(0.59)  -- fixup

	local line = AceGUI:Create("InteractiveSimpleGroup")
	line:SetLayout("Flow")
	line:SetFullWidth(true)
	line:AddChildren (itemlabel,eq)
	return line
end

local function mkbagsline (itemlink, bagtext, bagicon, slot, count)
	local itemlabel = mkitemlabel(itemlink,count)
	local bag = mklocationlabel(bagtext,bagicon)
	local slotw = mklocationlabel(slot)
	slotw:SetRelativeWidth(0.19)  -- fixup

	local line = AceGUI:Create("InteractiveSimpleGroup")
	line:SetLayout("Flow")
	line:SetFullWidth(true)
	line:AddChildren (itemlabel,bag,slotw)
	return line
end

--[[
local function mkvaultline (itemlink, bagtext, bagicon)
	local itemlabel = mkitemlabel(itemlink)
	local bag = mklocationlabel(bagtext,bagicon)
	local line = AceGUI:Create("SimpleGroup")
	line:SetLayout("Flow")
	line:SetFullWidth(true)
	line:AddChildren (itemlabel,bag)
	return line
end]]

local function mktotalsline (itemlink, count)
	local itemlabel = mkitemlabel(itemlink)   -- don't pass count here
	itemlabel:SetText(FMT.TOTAL_ITEM:format(itemlink,count))

	local line = AceGUI:Create("InteractiveSimpleGroup")
	line:SetLayout("Flow")
	line:SetFullWidth(true)
	line:AddChild(itemlabel)
	return line
end
local function mkalltotalsline (text)
	local label = AceGUI:Create("Label")
	if size == 'small' then
		label:SetFontObject(small_font_object)
		label:SetImageSize(small_button_image_size,small_button_image_size)
	elseif size == 'large' then
		label:SetFontObject(large_font_object)
		label:SetImageSize(large_button_image_size,large_button_image_size)
	end
	label:SetText(text)
	label:SetImage[[Interface\ICONS\Spell_ChargePositive]]
	label:SetFullWidth(true)
	local space = AceGUI:Create("Label")
	space:SetText[[ ]]
	space:SetFullWidth(true)
	return label, space
end

local function SpecialBottomButtonClicked (display, ev, button, key)
	display:Hide()
	addon.options.args.general.args.main.name = options_button_back
	LibStub("AceConfigDialog-3.0"):Open(nametag)
	LibStub("AceConfigDialog-3.0"):SelectGroup(nametag,key)
end

local function EditBox_OnFocus (display, event)
	display.editbox:HighlightText(0, event == "OnFocusGained" and -1 or 0)
end

local function EditBox_OnEnterPressed (display, event, text)
	addon:DoSearch(text)
end

function addon:DoSearch (text, redoing_p)
	assert(self.display and self.display:IsShown(), "searching without window")
	self.scroll:ReleaseChildren()

	if redoing_p then
		-- searchtargets preserved, ignore 'text'
	else
		text = text:trim()
		if #text < 1 then
			self:DPrint(2, "DoSearch given an empty string!")
			return
		end
		searchtargets = new(strsplit(';',text))
		local i = #searchtargets
		while i > 0 do
			local txt = searchtargets[i]
			if not txt then
				table.remove(searchtargets,i)
				i = #searchtargets  -- meh
			elseif #txt < 3 then
				self:Printf("%s: %s", txt, L["Search string is too short!"])
				table.remove(searchtargets,i)
				i = #searchtargets  -- meh
			else
				-- if they've linked an item to search, strip out the possibly-unique
				-- parts and just go with the item name
				searchtargets[i] = (txt:match("item:%d.*%[([^]]+)%]") or txt):trim():lower()
				i = i - 1
			end
		end
		if #searchtargets < 1 then
			del(searchtargets)
			return
		end
	end

	-- feedback to the user is good
	self.display:SetText(table.concat(searchtargets, "; "))

	self:Search(searchtargets)

	self:DPrint(3, "BuildWindow d=%s, d.frame=%s", tostring(self.display), tostring(self.display.frame))
	self:DPrint(3, "BuildWindow s=%s, s.frame=%s, s.sf=%s, s.c=%s",
		tostring(self.scroll), tostring(self.scroll.frame), tostring(self.scroll.scrollframe), tostring(self.scroll.content))

	self.display:PauseLayout()
	if cachefail then
		if redoing_p then
			-- This might both (a) never happen, and (b) not be true, since the
			-- cache does not get written to disk.
			self:Print(L["There are many missing items in your game cache.  You should revisit the banks for your alts."])
		else
			-- Listening for GET_ITEM_INFO_RECEIVED doesn't actually help us here.
			self.scroll:AddChild(mkblockline(L["Item cache error; the results below may be incomplete.  Automatically re-scanning in %d seconds..."]:format(cache_rescan_delay)))
			self:ScheduleTimer ("ReDoSearch", cache_rescan_delay)
		end
	else
		del(searchtargets)
	end
	if #results > 0 then
		local alltotal, totals = 0, new()
		for i,result in ipairs(results) do  -- dispatch table would be overkill
			local line
			if result.name then
				line = mkblockline (result.name)

			elseif result.whence == 'slots' then
				line = mkslotline (result.itemlink)
				totals[result.itemlink] = (totals[result.itemlink] or 0) + 1
				alltotal = alltotal + 1

			elseif result.whence == 'bags' then
				line = mkbagsline (result.itemlink, result.baglink, result.bagicon, result.slot, result.count)
				totals[result.itemlink] = (totals[result.itemlink] or 0) + (result.count or 1)
				alltotal = alltotal + (result.count or 1)

			elseif result.whence == 'vault' then
				line = mkbagsline (result.itemlink, result.tabname, result.tabicon, result.slot, result.count)
				totals[result.itemlink] = (totals[result.itemlink] or 0) + (result.count or 1)
				alltotal = alltotal + (result.count or 1)
			end

			self:DPrint(3, "Adding line %s", tostring(line.frame))
			self.scroll:AddChild(line)
		end
		self.scroll:AddChild(mkblockline(LIGHTYELLOW_FONT_COLOR_CODE..'++ '..L["Totals"]..' ++|r'))
		for itemlink,total in pairs(totals) do
			self.scroll:AddChild(mktotalsline(itemlink,total))
		end
		self.scroll:AddChildren(mkalltotalsline(FMT.TOTAL_ALL:format(alltotal)))
		del(totals)
	else
		self:DPrint(3, BROWSE_NO_RESULTS)
		self.scroll:AddChild(mkblockline(BROWSE_NO_RESULTS))
	end

	self.display:ResumeLayout()
	self.scroll:SetScroll(0)
end


function addon:ReDoSearch()
	if (not self.display) or (not self.display:IsShown()) then
		-- Actually, if the user has seen the "redoing in 7 sec" message and
		-- then closed the window, it's a safe bet that he doesn't want the
		-- thing reopening and stealing keyboard focus.
		--self:BuildWindow()
		return
	end
	self:DoSearch(nil,true)
end


function addon:BuildWindow()
	if self.display then
		self.display:Hide()
	end
	local display = AceGUI:Create("FrameWithInput")
	display:SetTitle(self.pretty_verstring)
	display:SetLayout("Fill")
	display:SetStatusTable{width=800}
	display:SetCallback("OnClose", function(_d)
		_G[nametag] = nil
		self.display = nil
		AceGUI:Release(_d)
		collectgarbage()
	end)
	self.display = display
	_G[nametag] = display  -- for Esc closing

	display:SetCallback("OnFocusGained", EditBox_OnFocus)
	display:SetCallback("OnFocusLost", EditBox_OnFocus)
	display:SetCallback("OnClick", SpecialBottomButtonClicked)
	display:SetCallback("OnEnterPressed", EditBox_OnEnterPressed)

	-- keys here match entries in addon.options.args
	display:AddButton ("general", L["Options"])  -- GAMEOPTIONS_MENU, maybe
	display:AddButton ("help", HELP_LABEL)

	local scroll = AceGUI:Create("ScrollFrame")
	scroll:SetLayout("List")
	self.scroll = scroll

	display:AddChild(scroll)
	display:ApplyStatus()
	display:SetFocus()
	display:Show()

	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	return display
end


-- Recycling functions yoinked from AceConfigDialog
do
	local pool = setmetatable({},{__mode="k"})
	function clear()
		wipe(pool)
	end
	function new(...)  -- slightly modified variant, takes optional initializers
		local t = next(pool)
		if t then
			pool[t] = nil
			for i = 1, select('#',...) do
				t[i] = select(i,...)
			end
			return t
		else
			return {...}
		end
	end
	function copy(t)
		local c = new()
		for k, v in pairs(t) do
			c[k] = v
		end
		return c
	end
	function del(t)
		wipe(t)
		pool[t] = true
	end
end


addon.error_exit = "ERROR:  Either none of the required DataStore library addons are loaded, or they are an unexpected version.  You need at least one (Inventory, Containers) to search.  If you think they're installed, please double-check the version requirements on *THIS* addon's project page at curse.com/curseforge.com/wowace.com as DataStore is extremely fragile."

addon.options.args.help.args.genH = {
	type = 'header', width = 'full', order = 1, name = [[How To Use]]}
addon.options.args.help.args.genT = {
	type = 'description', width = 'full', order = 2, name = ([===[
Type /rundown, optionally followed by a word/phrase or an item link, to open
the window and run a search.  Only toons which have logged on and visited the
bank will be searched.  You can search for multiple phrases at once by separating
them with a semicolon.

/rundown of the boar    [finds all random-type boar gear]

/rundown peaceb; silverl   [finds all Peacebloom and Silverleaf but leaves out things
like "Peacemaker" and "Silvermoon"]

Item names containing the string will be shown, along with their location.
If the item is in a bank or bag, its slot number will also be shown on the far
right side.  Slots start at 1 in the top-left corner of a bag, and count across the row
before going to the next row down.

If there are multiple items in a stack, the stack count will be shown on the
item icon or next to the item link, depending on the row size.
]===]):gsub("([^\n])\n([^\n])", "%1 %2")}


addon.options.args.help.args.optH = {
	type = 'header', width = 'full', order = 11, name = [[Options]]}
addon.options.args.help.args.optT = {
	type = 'description', width = 'full', order = 12, name = ([===[
Rundown supports standard Ace3 profiles, but the profiles only control the
options shown here; the item data is shared across all profiles.

The scope setting controls whether your searches are limited to toons on the current realm.
If you use cross-realm searching, then the realm name will be included in the character/guild
names shown on the display.

Bag numbering reminders are included due to an oddity in how Blizzard assigns
numbers to bags.  Purchased bag slots in your bank are numbered from left to right,
and the reminder includes a ">" sign.  Equipped bags next to your backpack are
numbered from right to left, and the reminder includes a "<" sign.

If you use a slash command followed by any of "-option", "-options", "-opt", or "-config",
then you'll get the options window instead of the normal one.  Ditto for "-?" and "-help" for
the instructions window.  A slash can be used in place of the dash. 
]===]):gsub("([^\n])\n([^\n])", "%1 %2")}


addon.options.args.help.args.histH = {
	type = 'header', width = 'full', order = 21, name = [[Episode... 4?]]}
addon.options.args.help.args.histT = {
	type = 'description', width = 'full', order = 22, name = ([===[
The original KC_Rundown/KC_Items was written by Kaelten back in mid-2005.  It used
the original Ace libraries (today called Ace1).  To the best of my knowledge it was
not formally updated for Ace2, so I kludged around the changes in my local copy.
KC_Rundown continued to work through Burning Crusade and Wrath, increasingly showing
cracks in the armor.

It finally stopped working altogether in Cataclysm.  Rather than trying to find a way
to make it (and Ace1!) work under the 4.x client, I wrote Rundown from scratch.  The
"Episode 4" reflects this history; the "New Hope" reflects my hope that eventually
somebody more competent than me will take over maintanence of this project.  :-)
]===]):gsub("([^\n])\n([^\n])", "%1 %2")}

-- vim:noet
